// Copyright (c) 2004-2008  Sundog Software, LLC. All rights reserved worldwide.

#include "Ephemeris.h"
#include <stdio.h>

#define PLANET_DELTA_DAYS (1.0 / 24.0)
#define PI 3.14159265
#define RADIANS(x)	((x) * (PI / 180.0))
#define DEGREES(x)	((x) * (180.0 / PI))

Ephemeris::Ephemeris()
{
    lastTime.SetYear(0);

    for (int i = 0; i < NUM_PLANETS; i++)
    {
        planets[i].lastEpochDaysCalculated = 0;
    }
}

void Ephemeris::Update(const LocalTime& time, const Location& location)
{
    bool timeChanged = false;
    bool locationChanged = false;

    if (time != lastTime)
    {
        timeChanged = true;
        lastTime = time;
    }

    if (location != lastLocation)
    {
        locationChanged = true;
        lastLocation = location;
    }

    if (timeChanged || locationChanged)
    {
        T = time.GetEpoch2000Centuries(true);
        Tuncorr = time.GetEpoch2000Centuries(false);
        epochDays = time.GetEpoch1990Days(false);

        Matrix3 Rx, Ry, Rz;
        Rx.FromRx(-0.1118 * T);
        Ry.FromRy(0.00972 * T);
        Rz.FromRz(-0.01118 * T);
        precession = Rz * (Ry * Rx);

        GMST = 4.894961 + 230121.675315 * Tuncorr; // radians
        LMST = GMST + RADIANS(location.GetLongitude()); // radians

        double latitude = RADIANS(location.GetLatitude());

        e = 0.409093 - 0.000227 * T;

        Ry.FromRy(-(latitude - PI / 2.0));
        Rz.FromRz(LMST);
        Rx.FromRx(-e);
        equatorialToHorizon = Ry * Rz * precession;
        eclipticToHorizon = Ry * Rz * Rx * precession;
        eclipticToEquatorial = Rx;

        equatorialToGeographic.FromRz(GMST);

        horizonToEquatorial = equatorialToHorizon.Transpose();
        horizonToGeographic = equatorialToGeographic * horizonToEquatorial;

        ComputeSunPosition();
        ComputeMoonPosition();
        ComputeEarthPosition();

        for (int i = 0; i < NUM_PLANETS; i++)
        {
            if (i != EARTH)
            {
                ComputePlanetPosition(i);
            }
        }
    }
}

Vector3 Ephemeris::ToCartesian(double r, double latitude, double longitude)
{
    Vector3 v;
    v.x = r * cos(longitude) * cos(latitude); // n
    v.y = r * sin(longitude) * cos(latitude); // e
    v.z = r * sin(latitude); // up

    return v;
}

double Ephemeris::Refract(double elevation)
{
    /*
    *    Refraction correction, degrees
    *        Zimmerman, John C.  1981.  Sun-pointing programs and their
    *            accuracy.
    *            SAND81-0761, Experimental Systems Operation Division 4721,
    *            Sandia National Laboratories, Albuquerque, NM.
    */

    //float prestemp;    /* temporary pressure/temperature correction */
    double refcor;      /* temporary refraction correction */
    double tanelev;     /* tangent of the solar elevation angle */

    /* If the sun is near zenith, the algorithm bombs; refraction near 0 */
    if ( elevation > RADIANS(85.0) )
        refcor = 0.0;

    /* Otherwise, we have refraction */
    else {
        tanelev = tan ( elevation );
        if ( elevation >= RADIANS(5.0) )
        {
            refcor  = 58.1 / tanelev -
                      0.07 / ( pow (tanelev,3) ) +
                      0.000086 / ( pow (tanelev,5) );
        }
        else if ( elevation >= RADIANS(-0.575) )
        {
            double degElev = DEGREES(elevation);
            refcor  = 1735.0 +
                      degElev * ( -518.2 + degElev * ( 103.4 +
                                                       degElev * ( -12.79 + degElev * 0.711 ) ) );
        }
        else
        {
            refcor  = -20.774 / tanelev;
        }
        //prestemp    =
        //    ( pdat->press * 283.0 ) / ( 1013.0 * ( 273.0 + pdat->temp ) );
        //refcor     *= prestemp / 3600.0;
        refcor /= 3600.0;
    }

    /* Refracted solar elevation angle */
    double out = elevation + RADIANS(refcor);

    return out;
}

void Ephemeris::ComputeSunPosition()
{
    double M = 6.24 + 628.302 * T;

    double longitude = 4.895048 + 628.331951 * T + (0.033417 - 0.000084 * T) * sin(M)
                       + 0.000351 * sin(2.0 * M);
    double latitude = 0;
    double geocentricDistance = 1.000140 - (0.016708 - 0.000042 * T) * cos(M) -
                                0.000141 * cos(2.0 * M); // AU's

    sunEclipticLongitude = longitude;
    sunEcl = ToCartesian(geocentricDistance, latitude, longitude);

    sunEq = eclipticToEquatorial * sunEcl;
    sunGeo = equatorialToGeographic * sunEq;

    // Oriented -x=n z=up y=east
    Vector3 tmp = eclipticToHorizon * sunEcl;
    sunHoriz.x = tmp.y; // x is east
    sunHoriz.y = tmp.z; // y is up
    sunHoriz.z = tmp.x; // -z is north

    // Account for atmospheric refraction.
    Vector3 tmp2 = sunHoriz;
    double R = tmp2.Length();
    tmp2.Normalize();
    double elev = asin(tmp2.y);
    elev = Refract(elev);
    sunHoriz.y = R * sin(elev);

}

static void InRange(double &in)
{
    while (in > 2.0 * PI) in -= 2.0 * PI;
    while (in < 0) in += 2.0 * PI;
}

void Ephemeris::ComputeMoonPosition()
{
    double lp = 3.8104 + 8399.7091 * T;
    double m = 6.2300 + 628.3019 * T;
    double f = 1.6280 + 8433.4663 * T;
    double mp = 2.3554 + 8328.6911 * T;
    double d = 5.1985 + 7771.3772 * T;

    double longitude =
        lp
        + 0.1098 * sin(mp)
        + 0.0222 * sin(2*d - mp)
        + 0.0115 * sin(2*d)
        + 0.0037 * sin(2*mp)
        - 0.0032 * sin(m)
        - 0.0020 * sin(2*f)
        + 0.0010 * sin(2*d - 2*mp)
        + 0.0010 * sin(2*d - m - mp)
        + 0.0009 * sin(2*d + mp)
        + 0.0008 * sin(2*d - m)
        + 0.0007 * sin(mp - m)
        - 0.0006 * sin(d)
        - 0.0005 * sin(m + mp);

    double latitude =
        + 0.0895 * sin(f)
        + 0.0049 * sin(mp + f)
        + 0.0048 * sin(mp - f)
        + 0.0030 * sin(2*d  - f)
        + 0.0010 * sin(2*d + f - mp)
        + 0.0008 * sin(2*d - f - mp)
        + 0.0006 * sin(2*d + f);

    double pip =
        + 0.016593
        + 0.000904 * cos(mp)
        + 0.000166 * cos(2*d - mp)
        + 0.000137 * cos(2*d)
        + 0.000049 * cos(2*mp)
        + 0.000015 * cos(2*d + mp)
        + 0.000009 * cos(2*d - m);

    double dMoon = 1.0 / pip; // earth radii

    moonEcl = ToCartesian(dMoon, latitude, longitude);

    moonEq = eclipticToEquatorial * moonEcl;
    moonGeo = equatorialToGeographic * moonEq;

    Vector3 tmp = (eclipticToHorizon * moonEcl);
    moonHoriz.x = tmp.y;
    moonHoriz.y = tmp.z;
    moonHoriz.z = tmp.x;

    InRange(longitude);
    InRange(sunEclipticLongitude);
    moonPhaseAngle = fabs(longitude - sunEclipticLongitude);
    InRange(moonPhaseAngle);
    moonPhase = 0.5 * (1.0 - cos(moonPhaseAngle));

    // Convert to +- 180 degrees
    if (moonPhase > PI)
    {
        moonPhase = PI - moonPhaseAngle;
    }

    moonDistance = (moonHoriz - Vector3(0, 0, 1)).Length() * 6378.137;
}

typedef struct OrbitalElements_S
{
    double period;
    double epochLongitude;
    double perihelionLongitude;
    double eccentricity;
    double semiMajorAxis;
    double inclination;
    double longitudeAscendingNode;
    double angularDiameter;
    double visualMagnitude;
} OrbitalElements;

OrbitalElements planetElements[NUM_PLANETS] =
{
    // Mercury
    {0.240852, RADIANS(60.750646), RADIANS(77.299833), 0.205633, 0.387099,
        RADIANS(7.004540), RADIANS(48.212740), 6.74, -0.42},

    // Venus
    {0.615211, RADIANS(88.455855), RADIANS(131.430236), 0.006778, 0.723332,
     RADIANS(3.394535), RADIANS(76.589820), 16.92, -4.40},

    // Earth
    {1.00004, RADIANS(99.403308), RADIANS(102.768413), 0.016713, 1.00000,
     0, 0, 0, 0},

    // Mars
    {1.880932, RADIANS(240.739474), RADIANS(335.874939), 0.093396, 1.523688,
     RADIANS(1.849736), RADIANS(49.480308), 9.36, -1.52},

    // Jupiter
    {11.863075, RADIANS(90.638185), RADIANS(14.170747), 0.048482, 5.202561,
     RADIANS(1.303613), RADIANS(100.353142), 196.74, -9.40},

    // Saturn
    {29.471362, RADIANS(287.690033), RADIANS(92.861407), 0.055581, 9.554747,
     RADIANS(2.488980), RADIANS(113.576139), 165.60, -8.88}
};

void Ephemeris::ComputeEarthPosition()
{
    double Np = ( (2.0 * PI) / 365.242191 ) * (epochDays / planetElements[EARTH].period);
    InRange(Np);

    double Mp = Np + planetElements[EARTH].epochLongitude - planetElements[EARTH].perihelionLongitude;

    L = Np + 2.0 * planetElements[EARTH].eccentricity * sin(Mp) +
        planetElements[EARTH].epochLongitude;

    InRange(L);

    double vp = L - planetElements[EARTH].perihelionLongitude;

    R = (planetElements[EARTH].semiMajorAxis * (1.0 - planetElements[EARTH].eccentricity
            * planetElements[EARTH].eccentricity)) / (1.0 + planetElements[EARTH].eccentricity
                    * cos(vp));
}

void Ephemeris::GetPlanetPosition(int planet, double& ra, double& dec, double& visualMagnitude) const
{
    if (planet < NUM_PLANETS)
    {
        ra = planets[planet].rightAscension;
        dec = planets[planet].declination;
        visualMagnitude = planets[planet].visualMagnitude;
    }
}

void Ephemeris::ComputePlanetPosition(int planet)
{
    if ((epochDays - planets[planet].lastEpochDaysCalculated) < PLANET_DELTA_DAYS)
    {
        return;
    }
    else
    {
        planets[planet].lastEpochDaysCalculated = epochDays;
    }

    double Np = ( (2.0 * PI) / 365.242191 ) * (epochDays / planetElements[planet].period);
    InRange(Np);

    double Mp = Np + planetElements[planet].epochLongitude - planetElements[planet].perihelionLongitude;

    double heliocentricLongitude = Np + 2.0 * planetElements[planet].eccentricity * sin(Mp) +
                                   planetElements[planet].epochLongitude;

    InRange(heliocentricLongitude);

    double vp = heliocentricLongitude - planetElements[planet].perihelionLongitude;

    double r = (planetElements[planet].semiMajorAxis * (1.0 - planetElements[planet].eccentricity
                * planetElements[planet].eccentricity)) / (1.0 + planetElements[planet].eccentricity
                        * cos(vp));

    double heliocentricLatitude = asin(sin(heliocentricLongitude -
                                           planetElements[planet].longitudeAscendingNode) * sin(planetElements[planet].inclination));

    InRange(heliocentricLatitude);

    double y = sin(heliocentricLongitude - planetElements[planet].longitudeAscendingNode) *
               cos(planetElements[planet].inclination);

    double x = cos(heliocentricLongitude - planetElements[planet].longitudeAscendingNode);

    double projectedHeliocentricLongitude = atan2(y, x) + planetElements[planet].longitudeAscendingNode;

    double projectedRadius = r * cos(heliocentricLatitude);

    double eclipticLongitude;

    if (planet > EARTH)
    {
        eclipticLongitude = atan( (R * sin(projectedHeliocentricLongitude - L))
                                  / (projectedRadius - R * cos(projectedHeliocentricLongitude - L)) )
                            + projectedHeliocentricLongitude;
    }
    else
    {
        eclipticLongitude = PI + L + atan( (projectedRadius * sin(L - projectedHeliocentricLongitude))
                                           / (R - projectedRadius * cos(L - projectedHeliocentricLongitude)) );
    }

    InRange(eclipticLongitude);

    double eclipticLatitude = atan( (projectedRadius * tan(heliocentricLatitude)
                                     * sin(eclipticLongitude - projectedHeliocentricLongitude))
                                    / (R * sin(projectedHeliocentricLongitude - L)));

    double ra = atan2( (sin(eclipticLongitude) * cos(e) - tan(eclipticLatitude) * sin(e))
                       , cos(eclipticLongitude));

    double dec = asin(sin(eclipticLatitude) * cos(e) + cos(eclipticLatitude) * sin(e)
                      * sin(eclipticLongitude));

    double dist2 = R * R + r * r - 2 * R * r * cos(heliocentricLongitude - L);
    double dist = sqrt(dist2);

    double d = eclipticLongitude - heliocentricLongitude;
    double phase = 0.5 * (1.0 + cos(d));

    double visualMagnitude;

    if (planet == VENUS)
    {
        d *= DEGREES(d);
        visualMagnitude = -4.34 + 5.0 * log10(r*dist) + 0.013 * d + 4.2E-7  * d*d*d;
    }
    else
    {
        visualMagnitude = 5.0 * log10( (r * dist) / sqrt(phase) ) + planetElements[planet].visualMagnitude;
    }

    planets[planet].rightAscension = ra;
    planets[planet].declination = dec;
    planets[planet].visualMagnitude = visualMagnitude;
}
